//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "Horizontal.h"
#include <inifiles.hpp>
#include <math.h>
#include "StringHelp.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
THorizontalForm *HorizontalForm;

/*
Program Copyright R.W.Williams (R.W.W) e-mail richardandtracy@yahoo.com

Permission is hereby granted to anyone who wants to, to use, distribute and
modify the program in any way they want free of charge for non-commercial use.
HOWEVER.... if the program is modified, the modifier MUST:-
1)  Acknowledge the R.W.W copyright material somewhere.
2)  EITHER send the modified program back to R.W.W for a cursory accuracy check
    OR state somewhere visible to the user that the program has been modified
    and the accuracy of the changes has not been verified.
3)  Not sell the modified program without discussions with R.W.W.
Commerial use must be discussed with R.W.W first.

There - those conditions aren't too onerous are they?
*/

//---------------------------------------------------------------------------
__fastcall THorizontalForm::THorizontalForm(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------

void __fastcall THorizontalForm::FormCreate(TObject *Sender)
{
    TIniFile *IniFile = new TIniFile(ExtractFilePath(Application->ExeName) + "Sundial.ini");
    LongDegrees->Text = IniFile->ReadString("Location","Long Degrees","0");
    LongMinutes->Text = IniFile->ReadString("Location","Long Minutes","0");
    LongSeconds->Text = IniFile->ReadString("Location","Long Seconds","0");
    LatDegrees->Text = IniFile->ReadString("Location","Lat Degrees","51");
    LatMinutes->Text = IniFile->ReadString("Location","Lat Minutes","32");
    LatSeconds->Text = IniFile->ReadString("Location","Lat Seconds","0.0");
    TimeZone->ItemIndex = IniFile->ReadInteger("Location","Time Zone Index",0) + 12;
    DaylightSaving->Checked = IniFile->ReadBool("Location","DST",false);
    FileName->Text = IniFile->ReadString("General","Last File",ExtractFilePath(Application->ExeName) + "Horizontal Sundial.dxf");
    delete IniFile;
}
//---------------------------------------------------------------------------

void __fastcall THorizontalForm::FormDestroy(TObject *Sender)
{
    TIniFile *IniFile = new TIniFile(ExtractFilePath(Application->ExeName) + "Sundial.ini");
    IniFile->WriteString("Location","Long Degrees",LongDegrees->Text);
    IniFile->WriteString("Location","Long Minutes",LongMinutes->Text);
    IniFile->WriteString("Location","Long Seconds",LongSeconds->Text);
    IniFile->WriteString("Location","Lat Degrees",LatDegrees->Text);
    IniFile->WriteString("Location","Lat Minutes",LatMinutes->Text);
    IniFile->WriteString("Location","Lat Seconds",LatSeconds->Text);
    IniFile->WriteInteger("Location","Time Zone Index",TimeZone->ItemIndex - 12);
    IniFile->WriteString("General","Last File",FileName->Text);
    IniFile->WriteBool("Location","DST",DaylightSaving->Checked);
    delete IniFile;
}
//---------------------------------------------------------------------------
void __fastcall THorizontalForm::BrowseButtonClick(TObject *Sender)
{
    SaveDialog1->InitialDir = ExtractFilePath(FileName->Text);
    SaveDialog1->FileName = FileName->Text;
    if (SaveDialog1->Execute()) {
        FileName->Text = SaveDialog1->FileName;
        }
}
//---------------------------------------------------------------------------
void __fastcall THorizontalForm::HelpButtonClick(TObject *Sender)
{
    //This is the help file for the dial drawing program
    AnsiString Message;
    Message = "This dial drawing program is used for horizontal sundials outside the Tropics.\n";
    Message += "\nEnter the longitude and latitude of the dial location with as great an accuracy as you can manage.\n";
    Message += "\tLongitude East: Degrees are to be preceeded by a - sign.\n\tLatitude South: Degrees are to be preceeded by a - sign.\n";
    Message += "\nSelect the time zone you want the dial to read, then select the output file for the dial drawing. This will be in .dxf format showing the hour lines and gnomon profile. The scale of the drawing is such that the gnomon length is 1.000 units. The drawing can be scaled in a CAD program once you have determined the gnomon length.\n";
    Message += "\nFinally, press the 'Go!' button.";

    Application->MessageBox(Message.c_str(),"Horizontal Sundial Design Program Help",MB_OK + MB_ICONINFORMATION);
}
//---------------------------------------------------------------------------
void __fastcall THorizontalForm::GoButtonClick(TObject *Sender)
{
    //This method does the calculation of the dial face dependant on all
    //the data entered so far - everything else is fluff!

    //All done within a try-catch block for safety.
try {
    Label3->Caption = "Starting Calculations";
    Label3->Refresh();

    DXFFile = new TStringList();
    WriteInitialLines();

    double latitude = ExtractAngle(LatDegrees, LatMinutes, LatSeconds);
    if (fabs(latitude) > 89.999) {
        Label3->Caption = "Latitude must be +89.999 to -89.999 degrees";
        return;
        }
    double longitude = ExtractAngle(LongDegrees, LongMinutes, LongSeconds);
    if (fabs(longitude) > 180.0) {
        Label3->Caption = "Longitude must be +180 to -180 degrees";
        return;
        }
    double DegreesToRadians = M_PI/180.0;

    //Draw the gnomon on the dial
    node1.x = 0.0;
    node1.y = 0.0;
    node2.x = 0.0;
    node2.y = 1.0;
    CreateLine(node1,node2);

    //Draw the gnomon profile.
    node1.x = 2.0;
    node1.y = 0.0;
    node2.x = 3.0;
    node2.y = 0.0;
    CreateLine(node1,node2);
    node2.x = 3.0;
    node2.y = tan(DegreesToRadians * (fabs(latitude)));
    CreateLine(node1,node2);
    node1.x = 3.0;
    node1.y = 0.0;
    CreateLine(node1,node2);
    node1.x = 2.50;
    node1.y = -0.2;
    CreateText(node1,"Gnomon",0.0);
    node1.y = -0.4;
    CreateText(node1,"Angle = " + HourMinSec(fabs(latitude)),0.0);
    node1.x = 2.0;
    node1.y = 0.0;
    CreateArc(node1,0.15,0.0,fabs(latitude));

    //Draw the inner perimiter of the dial (square)
    node1.x = -0.5;
    node1.y = 0.0;
    node2.x = -0.5;
    node2.y = 1.0;
    CreateLine(node1,node2);
    node1.x = 0.5;
    node1.y = 1.0;
    CreateLine(node1,node2);
    node2.x = 0.5;
    node2.y = 0.0;
    CreateLine(node1,node2);
    node1.x = 0.0;
    node1.y = -0.5;
    CreateText(node1,"Dial",0.0);
    node1.x = -0.5;
    node1.y = 0.0;
    node2.x = -0.5;
    node2.y = -0.25;
    CreateLine(node1,node2);
    node1.x = 0.5;
    node1.y = -0.25;
    CreateLine(node1,node2);
    node2.x = 0.5;
    node2.y = 0.0;
    CreateLine(node1,node2);

    //Draw the outer perimiter of the dial (square)
    node1.x = -0.6;
    node1.y = 0.0;
    node2.x = -0.6;
    node2.y = 1.1;
    CreateLine(node1,node2);
    node1.x = 0.6;
    node1.y = 1.1;
    CreateLine(node1,node2);
    node2.x = 0.6;
    node2.y = 0.0;
    CreateLine(node1,node2);
    node1.x = -0.6;
    node1.y = 0.0;
    node2.x = -0.6;
    node2.y = -0.35;
    CreateLine(node1,node2);
    node1.x = 0.6;
    node1.y = -0.35;
    CreateLine(node1,node2);
    node2.x = 0.6;
    node2.y = 0.0;
    CreateLine(node1,node2);

    //now to create the array of angles for local apparant time (LAT)
    DynamicArray <double> times;
    times.Length = 24;
    for (int i = 0; i < 24; i++) {
        times[i] = 270.0 - double(i) * 15.0;
        if (times[i] < 0.0) times[i] += 360;
        }

    //adjust array for difference between LAT and time zone
    double timezonelongitude;
    if (TimeZone->ItemIndex == TimeZone->Items->Count - 1) {
        timezonelongitude = longitude;
        } else {
        timezonelongitude = double(12 - TimeZone->ItemIndex) * 15.0;
        if (DaylightSaving->Checked) timezonelongitude -= 15.0;
        }
    double timedifference = longitude - timezonelongitude;
    for (int i = 0; i < 24; i++) {
        times[i] += timedifference;
        }

    //if south of equator, mirror about +/- 90 degree line
    if (latitude < 0.0) {
        for (int i = 0; i < 24; i++) {
            //while (times[i] < 0.0) times[i] += 360.0;
            double a, b;
            while (times[i] > 360.0) times[i] -= 360.0;
            a = times[i];
            if (times[i] <= 180.0) {
                times[i] = 180.0 - times[i];
                } else {
                times[i] = 180.0 + (360 - times[i]);
                }
            b = times[i];
            a = a/b;
            }
        }

    //Decide max/min hours to plot. This is different at different latitudes.
    //At the equator, never need more than 5am to 7pm. in the arctic circles
    //need to plot all the hours as the sun never sets in summer. In the
    //intermediate latitudes the sun may set at different times - and there is
    //no need to plot lines that can never be used.

    DynamicArray <bool> plot;
    plot.Length = 24;
    for (int i = 0; i < 24; i++) plot[i] = false;
    //plot LAT 5am-7pm
    double min = 195.0;
    double max = 345.0;

    //plot LAT 4am-8pm
    if (fabs(latitude) > 27.0) {
        min = 210;
        max = 330;
        }

    //plot LAT 3am-9pm
    if (fabs(latitude) > 46.0) {
        min = 225;
        max = 315;
        }

    //plot LAT 2am-10pm
    if (fabs(latitude) > 57.0) {
        min = 240;
        max = 300;
        }

    //plot LAT 1am-11pm
    if (fabs(latitude) > 65.0) {
        min = 255;
        max = 285;
        }

    //plot all hours in high latitudes
    if (fabs(latitude) > 66.0) {
        min = 271;
        max = 270;
        }

    for (int i = 0; i < 24; i++) {
        while (times[i] < 0.0) times[i] += 360;
        if (times[i] < min || times[i] > max) {
            plot[i] = true;
            }
        }

    //plot hour lines
    DynamicArray <double> angle;
    angle.Length = 24;
    node1.x = 0.0;
    node1.y = 0.0;
    for (int i = 0; i < 24; i++) {
        if (plot[i]) {
            //OK, now have a unit length line in the correct direction. Now to
            //check which part of the dial it intersects with.
            double textangle = 0.0;
            double lineangle;

            if (fmod(times[i],90.0) == 0.0) {
                lineangle = times[i];
                } else {
                lineangle = atan(tan((times[i]-90)*DegreesToRadians)*sin(fabs(latitude)*DegreesToRadians))/DegreesToRadians + 90;
                if (times[i] > 180.0) lineangle += 180.0;
                }

            angle[i] = lineangle;

            if (lineangle < 63.43495 || lineangle > 333.43495) {
                //intersects with right edge.
                node2.x = 0.5;
                node2.y = 0.5 * tan(lineangle * DegreesToRadians);
                } else {
                if (lineangle >= 63.43495 && lineangle < 116.56505) {
                    //top edge
                    node2.y = 1.0;
                    node2.x = 1.0 / tan(lineangle * DegreesToRadians);
                    } else {
                    if (lineangle >= 116.56505 && lineangle < 206.56505) {
                        //left edge
                        node2.x = -0.5;
                        node2.y = -0.5 * tan(lineangle * DegreesToRadians);
                        } else {
                        //bottom edge
                        node2.y = -0.25;
                        node2.x = -0.25 / tan(lineangle * DegreesToRadians);
                        }
                    }
                }

            //actually plot the line
            CreateLine(node1,node2);

            //sort the text angle & position out
            if (node2.x == 0.5) {
                textangle = 90;
                node2.x = 0.59;
                } else {
                if (node2.x == -0.5) {
                    textangle = 270;
                    node2.x = -0.59;
                    } else {
                    if (node2.y == 1.0) {
                        textangle = 180;
                        node2.y = 1.09;
                        } else {
                        node2.y = -0.34;
                        }
                    }
                }
            AnsiString text = AnsiString(i);
            if (i > 12) text = AnsiString(i-12);
            CreateText(node2,text,textangle);
            }
        }

    //plot table of angles to right of gnomon
    AnsiString characters;
    for (int i = 0; i < 24; i++) {
        if (plot[i]) {
            if (i < 13) {
                characters = AnsiString(i) + " am";
                } else {
                characters = AnsiString(i - 12) + " pm";
                }
            node1.x = 3.5;
            node1.y = 1.5 - 0.1 * double(i);
            CreateText(node1,characters,0.0);
            node1.x = 4.25;
            CreateText(node1,HourMinSec(angle[i]),0.0);
            }
        }

    //clear memory
    times.Length = 0;
    plot.Length = 0;
    angle.Length = 0;
    
    //finish dxf file
    WriteFinalLines();
    Label3->Caption = "Saving DXF File";
    Label3->Refresh();
    DXFFile->SaveToFile(FileName->Text);
    delete DXFFile;

    Label3->Caption = "Finished";
    }

catch (...) {
    //if there's a problem, stop the program bombing out and give a truly
    //unhelpful massage
    Label3->Caption = "An error ocurred - sorry!";
    }
}
//---------------------------------------------------------------------------

void __fastcall THorizontalForm::WriteInitialLines(void)
{
    DXFFile->Add("999");
    DXFFile->Add("DXF File Produced by RWW's Horizontal Sundial Program");
    DXFFile->Add("999");
    DXFFile->Add("Program Copyright R.W.Williams");
    DXFFile->Add("999");
    DXFFile->Add("e-mail: richardandtracy@yahoo.com");
    DXFFile->Add(" 0");
    DXFFile->Add("SECTION");
    DXFFile->Add(" 2");
    DXFFile->Add("ENTITIES");
}
//---------------------------------------------------------------------------

void __fastcall THorizontalForm::WriteFinalLines(void)
{
    //this is virtual so that it can be overridden by the wavefront obj conversion
    DXFFile->Add(" 0");
    DXFFile->Add("ENDSEC");
    DXFFile->Add(" 0");
    DXFFile->Add("EOF");
}
//---------------------------------------------------------------------------

void __fastcall THorizontalForm::CreateLine(nodalposition Node1, nodalposition Node2)
{
    //actually writes the line to the output file
    DXFFile->Add(" 0");
    DXFFile->Add("LINE");
    DXFFile->Add(" 8");
    DXFFile->Add("SUNDIAL");
    DXFFile->Add("62");
    DXFFile->Add("5");
    //end 1
    DXFFile->Add("10");
    DXFFile->Add(FloatToStrF(Node1.x,ffFixed,10,10));
    DXFFile->Add("20");
    DXFFile->Add(FloatToStrF(Node1.y,ffFixed,10,10));
    //end 2
    DXFFile->Add("11");
    DXFFile->Add(FloatToStrF(Node2.x,ffFixed,10,10));
    DXFFile->Add("21");
    DXFFile->Add(FloatToStrF(Node2.y,ffFixed,10,10));

}
//---------------------------------------------------------------------------

void __fastcall THorizontalForm::CreateArc(nodalposition Node1, double Radius, double StartAngle, double FinishAngle)
{
    //actually writes an arc to the output file
    DXFFile->Add(" 0");
    DXFFile->Add("ARC");
    DXFFile->Add(" 8");
    DXFFile->Add("SUNDIAL");
    DXFFile->Add("62");
    DXFFile->Add("5");
    //centre
    DXFFile->Add("10");
    DXFFile->Add(FloatToStrF(Node1.x,ffFixed,10,10));
    DXFFile->Add("20");
    DXFFile->Add(FloatToStrF(Node1.y,ffFixed,10,10));
    //radius
    DXFFile->Add("40");
    DXFFile->Add(FloatToStrF(Radius,ffFixed,10,10));
    //starting angle
    DXFFile->Add("50");
    DXFFile->Add(FloatToStrF(StartAngle,ffFixed,10,10));
    //finishing angle
    DXFFile->Add("51");
    DXFFile->Add(FloatToStrF(FinishAngle,ffFixed,10,10));

}
//---------------------------------------------------------------------------

void __fastcall THorizontalForm::CreateText(nodalposition Node1, AnsiString Text, double Angle)
{
    //actually writes the line to the output file
    //centred
    DXFFile->Add(" 0");
    DXFFile->Add("TEXT");
    DXFFile->Add(" 8");
    DXFFile->Add("SUNDIAL");
    DXFFile->Add("62");
    DXFFile->Add("5");
    DXFFile->Add("40");
    DXFFile->Add("0.075");
    DXFFile->Add("50");
    DXFFile->Add(AnsiString(Angle));
    DXFFile->Add("72");
    DXFFile->Add("1");
    DXFFile->Add("1");
    DXFFile->Add(Text);
    //location
    DXFFile->Add("10");
    DXFFile->Add(FloatToStrF(Node1.x,ffFixed,10,10));
    DXFFile->Add("20");
    DXFFile->Add(FloatToStrF(Node1.y,ffFixed,10,10));
    //Alignment point
    DXFFile->Add("11");
    DXFFile->Add(FloatToStrF(Node1.x,ffFixed,10,10));
    DXFFile->Add("21");
    DXFFile->Add(FloatToStrF(Node1.y,ffFixed,10,10));

}
//---------------------------------------------------------------------------


double __fastcall THorizontalForm::ExtractAngle(TEdit *Degrees, TEdit *Minutes, TEdit *Seconds)
{
    double angle = fabs(StrToFloat(Degrees->Text));
    bool negative = Degrees->Text.Pos("-");
    angle += fabs(StrToFloat(Minutes->Text)/60);
    angle += fabs(StrToFloat(Seconds->Text)/3600);

    if (negative) angle = angle * -1.0;

    return angle;

}
//---------------------------------------------------------------------------

